@@ -3,6 +3,8 @@
 
 #include "alloc/disk_groups.h"
 
+#include "data/reconcile/work.h"
+
 #include "init/dev.h"
 
 #include "sb/members.h"
@@ -86,6 +88,7 @@ static int bch2_sb_disk_groups_validate(struct bch_sb *sb, struct bch_sb_field *
 }
 
 static void bch2_sb_disk_groups_to_text(struct printbuf *out,
+					struct bch_fs *c,
 					struct bch_sb *sb,
 					struct bch_sb_field *f)
 {
@@ -193,7 +196,7 @@ const struct bch_devs_mask *bch2_target_to_mask(struct bch_fs *c, unsigned targe
 	}
 }
 
-bool bch2_dev_in_target(struct bch_fs *c, unsigned dev, unsigned target)
+bool bch2_dev_in_target_rcu(struct bch_fs *c, unsigned dev, unsigned target)
 {
 	if (dev == BCH_SB_MEMBER_INVALID)
 		return false;
@@ -449,7 +452,6 @@ int __bch2_dev_group_set(struct bch_fs *c, struct bch_dev *ca, const char *name)
 {
 	lockdep_assert_held(&c->sb_lock);
 
-
 	if (!strlen(name) || !strcmp(name, "none")) {
 		struct bch_member *mi = bch2_members_v2_get_mut(c->disk_sb.sb, ca->dev_idx);
 		SET_BCH_MEMBER_GROUP(mi, 0);
@@ -467,9 +469,19 @@ int __bch2_dev_group_set(struct bch_fs *c, struct bch_dev *ca, const char *name)
 
 int bch2_dev_group_set(struct bch_fs *c, struct bch_dev *ca, const char *name)
 {
-	guard(mutex)(&c->sb_lock);
-	return __bch2_dev_group_set(c, ca, name) ?:
-		bch2_write_super(c);
+	struct reconcile_scan s = { .type = RECONCILE_SCAN_pending };
+
+	try(bch2_set_reconcile_needs_scan(c, s, false));
+
+	/* bch2_reconcile_wakeup_pending goes here */
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
+	scoped_guard(mutex, &c->sb_lock) {
+		try(__bch2_dev_group_set(c, ca, name));
+		try(bch2_write_super(c));
+	}
+
+	try(bch2_set_reconcile_needs_scan(c, s, true));
+	return 0;
 }
 
 int bch2_opt_target_parse(struct bch_fs *c, const char *val, u64 *res,
