@@ -23,7 +23,7 @@ struct find_btree_nodes_worker {
 	struct bch_dev		*ca;
 };
 
-static void found_btree_node_to_text(struct printbuf *out, struct bch_fs *c, const struct found_btree_node *n)
+void bch2_found_btree_node_to_text(struct printbuf *out, struct bch_fs *c, const struct found_btree_node *n)
 {
 	bch2_btree_id_level_to_text(out, n->btree_id, n->level);
 	prt_printf(out, " seq=%u journal_seq=%llu cookie=%llx ",
@@ -34,24 +34,24 @@ static void found_btree_node_to_text(struct printbuf *out, struct bch_fs *c, con
 
 	if (n->range_updated)
 		prt_str(out, " range updated");
+	prt_newline(out);
 
 	guard(printbuf_indent)(out);
 	guard(printbuf_atomic)(out);
 	guard(rcu)();
 
 	for (unsigned i = 0; i < n->nr_ptrs; i++) {
-		prt_newline(out);
 		bch2_extent_ptr_to_text(out, c, n->ptrs + i);
+		prt_newline(out);
 	}
 }
 
-static void found_btree_nodes_to_text(struct printbuf *out, struct bch_fs *c, found_btree_nodes nodes)
+static void found_btree_nodes_to_text(struct printbuf *out, struct bch_fs *c,
+				      darray_found_btree_node nodes)
 {
 	guard(printbuf_indent)(out);
-	darray_for_each(nodes, i) {
-		found_btree_node_to_text(out, c, i);
-		prt_newline(out);
-	}
+	darray_for_each(nodes, i)
+		bch2_found_btree_node_to_text(out, c, i);
 }
 
 static void found_btree_node_to_key(struct bkey_i *k, const struct found_btree_node *f)
@@ -122,7 +122,7 @@ static const struct min_heap_callbacks found_btree_node_heap_cbs = {
 static void try_read_btree_node(struct find_btree_nodes *f, struct bch_dev *ca,
 				struct btree *b, struct bio *bio, u64 offset)
 {
-	struct bch_fs *c = container_of(f, struct bch_fs, found_btree_nodes);
+	struct bch_fs *c = container_of(f, struct bch_fs, btree.node_scan);
 	struct btree_node *bn = b->data;
 
 	bio_reset(bio, ca->disk_sb.bdev, REQ_OP_READ);
@@ -215,7 +215,7 @@ static void try_read_btree_node(struct find_btree_nodes *f, struct bch_dev *ca,
 static int read_btree_nodes_worker(void *p)
 {
 	struct find_btree_nodes_worker *w = p;
-	struct bch_fs *c = container_of(w->f, struct bch_fs, found_btree_nodes);
+	struct bch_fs *c = container_of(w->f, struct bch_fs, btree.node_scan);
 	struct bch_dev *ca = w->ca;
 	unsigned long last_print = jiffies;
 	struct btree *b = NULL;
@@ -235,30 +235,33 @@ static int read_btree_nodes_worker(void *p)
 		goto err;
 	}
 
+	u64 buckets_to_scan = 0;
 	for (u64 bucket = ca->mi.first_bucket; bucket < ca->mi.nbuckets; bucket++)
+		buckets_to_scan += c->sb.version_upgrade_complete < bcachefs_metadata_version_mi_btree_bitmap ||
+			bch2_dev_btree_bitmap_marked_sectors_any(ca, bucket_to_sector(ca, bucket), ca->mi.bucket_size);
+
+	u64 buckets_scanned = 0;
+	for (u64 bucket = ca->mi.first_bucket; bucket < ca->mi.nbuckets; bucket++) {
+		if (c->sb.version_upgrade_complete >= bcachefs_metadata_version_mi_btree_bitmap &&
+		    !bch2_dev_btree_bitmap_marked_sectors_any(ca, bucket_to_sector(ca, bucket), ca->mi.bucket_size))
+			continue;
+
 		for (unsigned bucket_offset = 0;
 		     bucket_offset + btree_sectors(c) <= ca->mi.bucket_size;
-		     bucket_offset += btree_sectors(c)) {
-			if (time_after(jiffies, last_print + HZ * 30)) {
-				u64 cur_sector = bucket * ca->mi.bucket_size + bucket_offset;
-				u64 end_sector = ca->mi.nbuckets * ca->mi.bucket_size;
+		     bucket_offset += btree_sectors(c))
+			try_read_btree_node(w->f, ca, b, bio, bucket_to_sector(ca, bucket) + bucket_offset);
 
-				bch_info(ca, "%s: %2u%% done", __func__,
-					 (unsigned) div64_u64(cur_sector * 100, end_sector));
-				last_print = jiffies;
-			}
+		buckets_scanned++;
 
-			u64 sector = bucket * ca->mi.bucket_size + bucket_offset;
-
-			if (c->sb.version_upgrade_complete >= bcachefs_metadata_version_mi_btree_bitmap &&
-			    !bch2_dev_btree_bitmap_marked_sectors(ca, sector, btree_sectors(c)))
-				continue;
-
-			try_read_btree_node(w->f, ca, b, bio, sector);
+		if (time_after(jiffies, last_print + HZ * 30)) {
+			bch_info_dev(ca, "%s: %2u%% done", __func__,
+				     (unsigned) div64_u64(buckets_scanned * 100, buckets_to_scan));
+			last_print = jiffies;
 		}
+	}
 err:
 	if (b)
-		__btree_node_data_free(b);
+		bch2_btree_node_data_free_locked(b);
 	kfree(b);
 	bio_put(bio);
 	enumerated_ref_put(&ca->io_ref[READ], BCH_DEV_READ_REF_btree_node_scan);
@@ -269,11 +272,10 @@ err:
 
 static int read_btree_nodes(struct find_btree_nodes *f)
 {
-	struct bch_fs *c = container_of(f, struct bch_fs, found_btree_nodes);
-	struct closure cl;
+	struct bch_fs *c = container_of(f, struct bch_fs, btree.node_scan);
 	int ret = 0;
 
-	closure_init_stack(&cl);
+	CLASS(closure_stack, cl)();
 	CLASS(printbuf, buf)();
 
 	prt_printf(&buf, "scanning for btree nodes on");
@@ -311,8 +313,10 @@ static int read_btree_nodes(struct find_btree_nodes *f)
 
 	bch_notice(c, "%s", buf.buf);
 err:
-	while (closure_sync_timeout(&cl, sysctl_hung_task_timeout_secs * HZ / 2))
+	/* Suppress hung task warning: */
+	while (closure_sync_timeout(&cl, HZ))
 		;
+
 	return f->ret ?: ret;
 }
 
@@ -326,7 +330,7 @@ static bool nodes_overlap(const struct found_btree_node *l,
 
 static int handle_overwrites(struct bch_fs *c,
 			     struct found_btree_node *l,
-			     found_btree_nodes *nodes_heap)
+			     darray_found_btree_node *nodes_heap)
 {
 	struct found_btree_node *r;
 
@@ -369,23 +373,21 @@ static int handle_overwrites(struct bch_fs *c,
 
 int bch2_scan_for_btree_nodes(struct bch_fs *c)
 {
-	struct find_btree_nodes *f = &c->found_btree_nodes;
+	struct find_btree_nodes *f = &c->btree.node_scan;
 	CLASS(printbuf, buf)();
-	found_btree_nodes nodes_heap = {};
+	CLASS(darray_found_btree_node, nodes_heap)();
 	size_t dst;
-	int ret = 0;
 
 	if (f->nodes.nr)
 		return 0;
 
-	mutex_init(&f->lock);
-
 	try(read_btree_nodes(f));
 
+	guard(mutex)(&f->lock);
+
 	if (!f->nodes.nr) {
 		bch_err(c, "%s: no btree nodes found", __func__);
-		ret = -EINVAL;
-		goto err;
+		return -EINVAL;
 	}
 
 	if (0 && c->opts.verbose) {
@@ -405,8 +407,7 @@ int bch2_scan_for_btree_nodes(struct bch_fs *c)
 		    prev->cookie == i->cookie) {
 			if (prev->nr_ptrs == ARRAY_SIZE(prev->ptrs)) {
 				bch_err(c, "%s: found too many replicas for btree node", __func__);
-				ret = -EINVAL;
-				goto err;
+				return -EINVAL;
 			}
 			prev->ptrs[prev->nr_ptrs++] = i->ptrs[0];
 		} else {
@@ -429,33 +430,27 @@ int bch2_scan_for_btree_nodes(struct bch_fs *c)
 	{
 		/* darray must have same layout as a heap */
 		min_heap_char real_heap;
-		BUILD_BUG_ON(sizeof(nodes_heap.nr)	!= sizeof(real_heap.nr));
-		BUILD_BUG_ON(sizeof(nodes_heap.size)	!= sizeof(real_heap.size));
-		BUILD_BUG_ON(offsetof(found_btree_nodes, nr)	!= offsetof(min_heap_char, nr));
-		BUILD_BUG_ON(offsetof(found_btree_nodes, size)	!= offsetof(min_heap_char, size));
+		BUILD_BUG_ON(sizeof(nodes_heap.nr)			!= sizeof(real_heap.nr));
+		BUILD_BUG_ON(sizeof(nodes_heap.size)			!= sizeof(real_heap.size));
+		BUILD_BUG_ON(offsetof(darray_found_btree_node, nr)	!= offsetof(min_heap_char, nr));
+		BUILD_BUG_ON(offsetof(darray_found_btree_node, size)	!= offsetof(min_heap_char, size));
 	}
 
 	min_heapify_all(&nodes_heap, &found_btree_node_heap_cbs, NULL);
 
 	if (nodes_heap.nr) {
-		ret = darray_push(&f->nodes, *min_heap_peek(&nodes_heap));
-		if (ret)
-			goto err;
+		try(darray_push(&f->nodes, *min_heap_peek(&nodes_heap)));
 
 		min_heap_pop(&nodes_heap, &found_btree_node_heap_cbs, NULL);
 	}
 
 	while (true) {
-		ret = handle_overwrites(c, &darray_last(f->nodes), &nodes_heap);
-		if (ret)
-			goto err;
+		try(handle_overwrites(c, &darray_last(f->nodes), &nodes_heap));
 
 		if (!nodes_heap.nr)
 			break;
 
-		ret = darray_push(&f->nodes, *min_heap_peek(&nodes_heap));
-		if (ret)
-			goto err;
+		try(darray_push(&f->nodes, *min_heap_peek(&nodes_heap)));
 
 		min_heap_pop(&nodes_heap, &found_btree_node_heap_cbs, NULL);
 	}
@@ -473,9 +468,7 @@ int bch2_scan_for_btree_nodes(struct bch_fs *c)
 	}
 
 	eytzinger0_sort(f->nodes.data, f->nodes.nr, sizeof(f->nodes.data[0]), found_btree_node_cmp_pos, NULL);
-err:
-	darray_exit(&nodes_heap);
-	return ret;
+	return 0;
 }
 
 static int found_btree_node_range_start_cmp(const void *_l, const void *_r)
@@ -500,7 +493,7 @@ static int found_btree_node_range_start_cmp(const void *_l, const void *_r)
 
 bool bch2_btree_node_is_stale(struct bch_fs *c, struct btree *b)
 {
-	struct find_btree_nodes *f = &c->found_btree_nodes;
+	struct find_btree_nodes *f = &c->btree.node_scan;
 
 	struct found_btree_node search = {
 		.btree_id	= b->c.btree_id,
@@ -515,9 +508,9 @@ bool bch2_btree_node_is_stale(struct bch_fs *c, struct btree *b)
 	return false;
 }
 
-int bch2_btree_has_scanned_nodes(struct bch_fs *c, enum btree_id btree)
+int bch2_btree_has_scanned_nodes(struct bch_fs *c, enum btree_id btree, struct printbuf *out)
 {
-	try(bch2_run_print_explicit_recovery_pass(c, BCH_RECOVERY_PASS_scan_for_btree_nodes));
+	try(bch2_run_explicit_recovery_pass(c, out, BCH_RECOVERY_PASS_scan_for_btree_nodes, 0));
 
 	struct found_btree_node search = {
 		.btree_id	= btree,
@@ -526,18 +519,19 @@ int bch2_btree_has_scanned_nodes(struct bch_fs *c, enum btree_id btree)
 		.max_key	= SPOS_MAX,
 	};
 
-	for_each_found_btree_node_in_range(&c->found_btree_nodes, search, idx)
+	for_each_found_btree_node_in_range(&c->btree.node_scan, search, idx)
 		return true;
 	return false;
 }
 
 int bch2_get_scanned_nodes(struct bch_fs *c, enum btree_id btree,
-			   unsigned level, struct bpos node_min, struct bpos node_max)
+			   unsigned level, struct bpos node_min, struct bpos node_max,
+			   struct printbuf *out, size_t *nodes_found)
 {
 	if (!btree_id_recovers_from_scan(btree))
 		return 0;
 
-	try(bch2_run_print_explicit_recovery_pass(c, BCH_RECOVERY_PASS_scan_for_btree_nodes));
+	try(bch2_run_explicit_recovery_pass(c, out, BCH_RECOVERY_PASS_scan_for_btree_nodes, 0));
 
 	if (c->opts.verbose) {
 		CLASS(printbuf, buf)();
@@ -559,7 +553,7 @@ int bch2_get_scanned_nodes(struct bch_fs *c, enum btree_id btree,
 		.max_key	= node_max,
 	};
 
-	struct find_btree_nodes *f = &c->found_btree_nodes;
+	struct find_btree_nodes *f = &c->btree.node_scan;
 	for_each_found_btree_node_in_range(f, search, idx) {
 		struct found_btree_node n = f->nodes.data[idx];
 
@@ -573,12 +567,6 @@ int bch2_get_scanned_nodes(struct bch_fs *c, enum btree_id btree,
 
 		found_btree_node_to_key(&tmp.k, &n);
 
-		if (c->opts.verbose) {
-			CLASS(printbuf, buf)();
-			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&tmp.k));
-			bch_verbose(c, "%s(): recovering %s", __func__, buf.buf);
-		}
-
 		BUG_ON(bch2_bkey_validate(c, bkey_i_to_s_c(&tmp.k),
 					  (struct bkey_validate_context) {
 						.from	= BKEY_VALIDATE_btree_node,
@@ -586,9 +574,27 @@ int bch2_get_scanned_nodes(struct bch_fs *c, enum btree_id btree,
 						.btree	= btree,
 					  }));
 
+		if (!*nodes_found) {
+			prt_printf(out, "recovering from btree node scan at ");
+			bch2_btree_id_level_to_text(out, btree, level);
+			prt_newline(out);
+			printbuf_indent_add(out, 2);
+		}
+
+		*nodes_found += 1;
+
+		if (*nodes_found < 10) {
+			bch2_bkey_val_to_text(out, c, bkey_i_to_s_c(&tmp.k));
+			prt_newline(out);
+		} else if (*nodes_found == 10)
+			prt_printf(out, "<many>\n");
+
 		try(bch2_journal_key_insert(c, btree, level + 1, &tmp.k));
 	}
 
+	if (*nodes_found)
+		printbuf_indent_sub(out, 2);
+
 	return 0;
 }
 
@@ -596,3 +602,8 @@ void bch2_find_btree_nodes_exit(struct find_btree_nodes *f)
 {
 	darray_exit(&f->nodes);
 }
+
+void bch2_find_btree_nodes_init(struct find_btree_nodes *f)
+{
+	mutex_init(&f->lock);
+}
