@@ -16,6 +16,8 @@
 
 #include "init/error.h"
 
+#include "sb/counters.h"
+
 #include "snapshots/snapshot.h"
 
 #include <linux/string_helpers.h>
@@ -241,7 +243,6 @@ static int bch2_trans_update_extent(struct btree_trans *trans,
 
 	CLASS(btree_iter, iter)(trans, btree_id, bkey_start_pos(&insert->k),
 				BTREE_ITER_intent|
-				BTREE_ITER_with_updates|
 				BTREE_ITER_not_extents|
 				BTREE_ITER_nofilter_whiteouts);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_max(&iter, POS(insert->k.p.inode, U64_MAX)));
@@ -314,6 +315,13 @@ __btree_trans_update_by_path(struct btree_trans *trans,
 	EBUG_ON(!path->should_be_locked);
 	EBUG_ON(trans->nr_updates >= trans->nr_paths);
 	EBUG_ON(!bpos_eq(k->k.p, path->pos));
+	EBUG_ON(!path->level &&
+		btree_type_has_snapshots(path->btree_id) &&
+		!bkey_deleted(&k->k) &&
+		test_bit(JOURNAL_replay_done, &c->journal.flags) &&
+		!bch2_snapshot_exists(c, k->k.p.snapshot));
+
+	trans->has_interior_updates |= path->level != 0;
 
 	n = (struct btree_insert_entry) {
 		.flags		= flags,
@@ -374,7 +382,12 @@ __btree_trans_update_by_path(struct btree_trans *trans,
 
 	__btree_path_get(trans, trans->paths + i->path, true);
 
-	trace_update_by_path(trans, path, i, overwrite);
+	event_trace(c, update_by_path, buf, ({
+		prt_printf(&buf, "%s overwrite %u\n", trans->fn, overwrite);
+		bch2_btree_path_to_text_short(&buf, trans, path_idx, path);
+		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(k));
+	}));
+
 	return i;
 }
 
@@ -416,6 +429,12 @@ static noinline int flush_new_cached_update(struct btree_trans *trans,
 	return 0;
 }
 
+static inline bool key_cache_needs_flush(struct btree_path *path)
+{
+	struct bkey_cached *ck = (void *) path->l[0].b;
+	return ck->needs_immediate_flush;
+}
+
 static int __must_check
 bch2_trans_update_by_path(struct btree_trans *trans, btree_path_idx_t path_idx,
 			  struct bkey_i *k, enum btree_iter_update_trigger_flags flags,
@@ -430,7 +449,9 @@ bch2_trans_update_by_path(struct btree_trans *trans, btree_path_idx_t path_idx,
 	 * the key cache - but the key has to exist in the btree for that to
 	 * work:
 	 */
-	return i->cached && (!i->old_btree_u64s || bkey_deleted(&k->k))
+	return i->cached && (!i->old_btree_u64s ||
+			     bkey_deleted(&k->k) ||
+			     key_cache_needs_flush(trans->paths + path_idx))
 		? flush_new_cached_update(trans, i, flags, ip)
 		: 0;
 }
@@ -460,7 +481,7 @@ static noinline int bch2_trans_update_get_key_cache(struct btree_trans *trans,
 		struct bkey_cached *ck = (void *) trans->paths[iter->key_cache_path].l[0].b;
 
 		if (test_bit(BKEY_CACHED_DIRTY, &ck->flags)) {
-			trace_and_count(trans->c, trans_restart_key_cache_raced, trans, _RET_IP_);
+			event_inc_trace(trans->c, trans_restart_key_cache_raced, buf, prt_str(&buf, trans->fn));
 			return btree_trans_restart(trans, BCH_ERR_transaction_restart_key_cache_raced);
 		}
 
@@ -499,7 +520,7 @@ int __must_check bch2_trans_update_ip(struct btree_trans *trans, struct btree_it
 	if (!(flags & BTREE_UPDATE_key_cache_reclaim) &&
 	    !path->cached &&
 	    !path->level &&
-	    btree_id_cached(trans->c, path->btree_id)) {
+	    btree_id_cached(path->btree_id)) {
 		try(bch2_trans_update_get_key_cache(trans, iter, path));
 
 		path_idx = iter->key_cache_path;
@@ -552,7 +573,7 @@ void *__bch2_trans_subbuf_alloc(struct btree_trans *trans,
 int bch2_bkey_get_empty_slot(struct btree_trans *trans, struct btree_iter *iter,
 			     enum btree_id btree, struct bpos start, struct bpos end)
 {
-	bch2_trans_iter_init(trans, iter, btree, end, BTREE_ITER_intent|BTREE_ITER_with_updates);
+	bch2_trans_iter_init(trans, iter, btree, end, BTREE_ITER_intent);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_prev(iter));
 
 	if (bpos_lt(iter->pos, start))
