@@ -1,6 +1,8 @@
 // SPDX-License-Identifier: GPL-2.0
 #include "bcachefs.h"
 
+#include "btree/iter.h"
+
 #include "data/checksum.h"
 #include "data/compress.h"
 #include "data/extents.h"
@@ -14,6 +16,12 @@
 #include <linux/zlib.h>
 #include <linux/zstd.h>
 
+#include <linux/module.h>
+
+static bool bch2_verify_compress = IS_ENABLED(CONFIG_BCACHEFS_DEBUG);
+module_param_named(verify_compress, bch2_verify_compress, bool, 0644);
+MODULE_PARM_DESC(verify_compress, "Decompress data immediately after compressing, and verify the result");
+
 static inline enum bch_compression_opts bch2_compression_type_to_opt(enum bch_compression_type type)
 {
 	switch (type) {
@@ -182,14 +190,10 @@ static inline void zlib_set_workspace(z_stream *strm, void *workspace)
 #endif
 }
 
-static int __bio_uncompress(struct bch_fs *c, struct bio *src,
-			    void *dst_data, struct bch_extent_crc_unpacked crc)
+static int buf_uncompress(struct bch_fs *c,
+			  void *dst, void *src,
+			  struct bch_extent_crc_unpacked crc)
 {
-	size_t src_len = src->bi_iter.bi_size;
-	size_t dst_len = crc.uncompressed_size << 9;
-	void *workspace;
-	int ret2;
-
 	enum bch_compression_opts opt = bch2_compression_type_to_opt(crc.compression_type);
 	mempool_t *workspace_pool = &c->compress.workspace[opt];
 	if (unlikely(!mempool_initialized(workspace_pool))) {
@@ -201,54 +205,61 @@ static int __bio_uncompress(struct bch_fs *c, struct bio *src,
 			return bch_err_throw(c, compression_workspace_not_initialized);
 	}
 
-	struct bbuf src_data __cleanup(bbuf_exit) = bio_map_or_bounce(c, src, READ);
+	size_t src_len = crc.compressed_size << 9;
+	size_t dst_len = crc.uncompressed_size << 9;
 
 	switch (crc.compression_type) {
 	case BCH_COMPRESSION_TYPE_lz4_old:
-	case BCH_COMPRESSION_TYPE_lz4:
-		ret2 = LZ4_decompress_safe_partial(src_data.b, dst_data,
-						   src_len, dst_len, dst_len);
-		if (ret2 != dst_len)
+	case BCH_COMPRESSION_TYPE_lz4: {
+		int ret = LZ4_decompress_safe_partial(src, dst, src_len, dst_len, dst_len);
+		if (ret != dst_len)
 			return bch_err_throw(c, decompress_lz4);
 		break;
+	}
 	case BCH_COMPRESSION_TYPE_gzip: {
 		z_stream strm = {
-			.next_in	= src_data.b,
+			.next_in	= src,
 			.avail_in	= src_len,
-			.next_out	= dst_data,
+			.next_out	= dst,
 			.avail_out	= dst_len,
 		};
 
-		workspace = mempool_alloc(workspace_pool, GFP_NOFS);
+		void *workspace = mempool_alloc(workspace_pool, GFP_NOFS);
 
 		zlib_set_workspace(&strm, workspace);
 		zlib_inflateInit2(&strm, -MAX_WBITS);
-		ret2 = zlib_inflate(&strm, Z_FINISH);
+		int ret = zlib_inflate(&strm, Z_FINISH);
 
 		mempool_free(workspace, workspace_pool);
 
-		if (ret2 != Z_STREAM_END)
+		if (ret != Z_STREAM_END)
 			return bch_err_throw(c, decompress_gzip);
 		break;
 	}
 	case BCH_COMPRESSION_TYPE_zstd: {
 		ZSTD_DCtx *ctx;
-		size_t real_src_len = le32_to_cpup(src_data.b);
+		size_t real_src_len = le32_to_cpup(src);
 
 		if (real_src_len > src_len - 4)
 			return bch_err_throw(c, decompress_zstd_src_len_bad);
 
-		workspace = mempool_alloc(workspace_pool, GFP_NOFS);
+		void *workspace = mempool_alloc(workspace_pool, GFP_NOFS);
 		ctx = zstd_init_dctx(workspace, zstd_dctx_workspace_bound());
 
-		ret2 = zstd_decompress_dctx(ctx,
-				dst_data,	dst_len,
-				src_data.b + 4, real_src_len);
+		size_t ret = zstd_decompress_dctx(ctx,
+				dst,	dst_len,
+				src + 4, real_src_len);
 
 		mempool_free(workspace, workspace_pool);
 
-		if (ret2 != dst_len)
-			return bch_err_throw(c, decompress_zstd);
+		if (ret != dst_len) {
+			if (zstd_is_error(ret))
+				return __bch2_err_throw(c, zstd_err_to_bch_err(zstd_get_error_code(ret)));
+			else {
+				bch_err(c, "zstd decompress error: expected %zu bytes, got %zu", dst_len, ret);
+				return bch_err_throw(c, decompress_zstd_size_mismatch);
+			}
+		}
 		break;
 	}
 	default:
@@ -268,6 +279,7 @@ int bch2_bio_uncompress_inplace(struct bch_write_op *op,
 	/* bio must own its pages: */
 	BUG_ON(!bio->bi_vcnt);
 	BUG_ON(DIV_ROUND_UP(crc->live_size, PAGE_SECTORS) > bio->bi_max_vecs);
+	BUG_ON(bio->bi_iter.bi_size != crc->compressed_size << 9);
 
 	if (crc->uncompressed_size << 9	> c->opts.encoded_extent_max) {
 		bch2_write_op_error(op, false, op->pos.offset,
@@ -276,9 +288,10 @@ int bch2_bio_uncompress_inplace(struct bch_write_op *op,
 		return bch_err_throw(c, decompress_exceeded_max_encoded_extent);
 	}
 
-	struct bbuf data __cleanup(bbuf_exit) = __bounce_alloc(c, dst_len, WRITE);
+	struct bbuf dst_buf __cleanup(bbuf_exit) = __bounce_alloc(c, dst_len, WRITE);
+	struct bbuf src_buf __cleanup(bbuf_exit) = bio_map_or_bounce(c, bio, READ);
 
-	int ret = __bio_uncompress(c, bio, data.b, *crc);
+	int ret = buf_uncompress(c, dst_buf.b, src_buf.b, *crc);
 	if (c->opts.no_data_io)
 		ret = 0;
 	if (ret) {
@@ -292,7 +305,7 @@ int bch2_bio_uncompress_inplace(struct bch_write_op *op,
 	 */
 	bio->bi_iter.bi_size = crc->live_size << 9;
 
-	memcpy_to_bio(bio, bio->bi_iter, data.b + (crc->offset << 9));
+	memcpy_to_bio(bio, bio->bi_iter, dst_buf.b + (crc->offset << 9));
 
 	crc->csum_type		= 0;
 	crc->compression_type	= 0;
@@ -304,24 +317,28 @@ int bch2_bio_uncompress_inplace(struct bch_write_op *op,
 }
 
 int bch2_bio_uncompress(struct bch_fs *c, struct bio *src,
-		       struct bio *dst, struct bvec_iter dst_iter,
-		       struct bch_extent_crc_unpacked crc)
+			struct bio *dst, struct bvec_iter dst_iter,
+			struct bch_extent_crc_unpacked crc)
 {
 	size_t dst_len = crc.uncompressed_size << 9;
 
+	BUG_ON(dst_iter.bi_size + (crc.offset << 9) > dst_len);
+	BUG_ON(src->bi_iter.bi_size != crc.compressed_size << 9);
+
 	if (crc.uncompressed_size << 9	> c->opts.encoded_extent_max ||
 	    crc.compressed_size << 9	> c->opts.encoded_extent_max)
 		return bch_err_throw(c, decompress_exceeded_max_encoded_extent);
 
-	struct bbuf dst_data __cleanup(bbuf_exit) = dst_len == dst_iter.bi_size
+	struct bbuf dst_buf __cleanup(bbuf_exit) = dst_len == dst_iter.bi_size
 		? __bio_map_or_bounce(c, dst, dst_iter, WRITE)
 		: __bounce_alloc(c, dst_len, WRITE);
+	struct bbuf src_buf __cleanup(bbuf_exit) = bio_map_or_bounce(c, src, READ);
 
-	try(__bio_uncompress(c, src, dst_data.b, crc));
+	try(buf_uncompress(c, dst_buf.b, src_buf.b, crc));
 
-	if (dst_data.type != BB_none &&
-	    dst_data.type != BB_vmap)
-		memcpy_to_bio(dst, dst_iter, dst_data.b + (crc.offset << 9));
+	if (dst_buf.type != BB_none &&
+	    dst_buf.type != BB_vmap)
+		memcpy_to_bio(dst, dst_iter, dst_buf.b + (crc.offset << 9));
 	return 0;
 }
 
@@ -334,6 +351,9 @@ static int attempt_compress(struct bch_fs *c,
 	enum bch_compression_type compression_type =
 		__bch2_compression_opt_to_type[compression.type];
 
+	BUG_ON(src_len & 511);
+	BUG_ON(dst_len & 511);
+
 	switch (compression_type) {
 	case BCH_COMPRESSION_TYPE_lz4:
 		if (compression.level < LZ4HC_MIN_CLEVEL) {
@@ -416,11 +436,19 @@ static int attempt_compress(struct bch_fs *c,
 	}
 }
 
-static unsigned __bio_compress(struct bch_fs *c,
-			       struct bio *dst, size_t *dst_len,
-			       struct bio *src, size_t *src_len,
-			       union bch_compression_opt compression)
+static unsigned bch2_compress(struct bch_fs *c,
+			      void *dst, size_t *dst_len,
+			      void *src, size_t *src_len,
+			      unsigned compression_opt,
+			      struct bpos write_pos)
 {
+	union bch_compression_opt compression =
+		(union bch_compression_opt) { .value = compression_opt };
+
+	/* If it's only one block, don't bother trying to compress: */
+	if (*src_len <= c->opts.block_size)
+		return BCH_COMPRESSION_TYPE_incompressible;
+
 	enum bch_compression_type compression_type =
 		__bch2_compression_opt_to_type[compression.type];
 	int ret = 0;
@@ -441,18 +469,8 @@ static unsigned __bio_compress(struct bch_fs *c,
 		}
 	}
 
-	/* If it's only one block, don't bother trying to compress: */
-	if (src->bi_iter.bi_size <= c->opts.block_size)
-		return BCH_COMPRESSION_TYPE_incompressible;
-
-	struct bbuf dst_data __cleanup(bbuf_exit) = bio_map_or_bounce(c, dst, WRITE);
-	struct bbuf src_data __cleanup(bbuf_exit) = bio_map_or_bounce(c, src, READ);
-
 	void *workspace = mempool_alloc(workspace_pool, GFP_NOFS);
 
-	*src_len = src->bi_iter.bi_size;
-	*dst_len = dst->bi_iter.bi_size;
-
 	/*
 	 * XXX: this algorithm sucks when the compression code doesn't tell us
 	 * how much would fit, like LZ4 does:
@@ -464,8 +482,8 @@ static unsigned __bio_compress(struct bch_fs *c,
 		}
 
 		ret = attempt_compress(c, workspace,
-				       dst_data.b,	*dst_len,
-				       src_data.b,	*src_len,
+				       dst, *dst_len,
+				       src, *src_len,
 				       compression);
 		if (ret > 0) {
 			*dst_len = ret;
@@ -502,15 +520,34 @@ static unsigned __bio_compress(struct bch_fs *c,
 
 	unsigned pad = round_up(*dst_len, block_bytes(c)) - *dst_len;
 
-	memset(dst_data.b + *dst_len, 0, pad);
+	memset(dst + *dst_len, 0, pad);
 	*dst_len += pad;
 
-	if (dst_data.type != BB_none &&
-	    dst_data.type != BB_vmap)
-		memcpy_to_bio(dst, dst->bi_iter, dst_data.b);
+	if (unlikely(bch2_verify_compress)) {
+		struct bch_extent_crc_unpacked crc = {
+			.compressed_size	= *dst_len >> 9,
+			.uncompressed_size	= *src_len >> 9,
+			.compression_type	= compression_type,
+		};
+
+		struct bbuf verify __cleanup(bbuf_exit) = __bounce_alloc(c, *src_len, WRITE);
+		ret = buf_uncompress(c, verify.b, dst, crc);
+		BUG_ON(ret);
+
+		if (memcmp(verify.b, src, *src_len)) {
+			CLASS(bch_log_msg, msg)(c);
+			prt_printf(&msg.m, "Decompressing compressed data did not produce the same result (%s)",
+				   __bch2_compression_types[compression_type]);
+
+			CLASS(btree_trans, trans)(c);
+			bch2_inum_offset_err_msg_trans(trans, &msg.m, 0, write_pos);
+			prt_printf(&msg.m, " len %zu\n", *src_len);
+
+			msg.m.suppress = bch2_count_fsck_err(c, compression_error, &msg.m);
+			return BCH_COMPRESSION_TYPE_incompressible;
+		}
+	}
 
-	BUG_ON(!*dst_len || *dst_len > dst->bi_iter.bi_size);
-	BUG_ON(!*src_len || *src_len > src->bi_iter.bi_size);
 	BUG_ON(*dst_len & (block_bytes(c) - 1));
 	BUG_ON(*src_len & (block_bytes(c) - 1));
 	return compression_type;
@@ -519,24 +556,45 @@ static unsigned __bio_compress(struct bch_fs *c,
 unsigned bch2_bio_compress(struct bch_fs *c,
 			   struct bio *dst, size_t *dst_len,
 			   struct bio *src, size_t *src_len,
-			   unsigned compression_opt)
+			   unsigned compression_opt,
+			   struct bpos write_pos,
+			   bool bounce_source)
 {
-	unsigned orig_dst = dst->bi_iter.bi_size;
-	unsigned orig_src = src->bi_iter.bi_size;
-	unsigned compression_type;
-
 	/* Don't consume more than BCH_ENCODED_EXTENT_MAX from @src: */
-	src->bi_iter.bi_size = min_t(unsigned, src->bi_iter.bi_size,
-				     c->opts.encoded_extent_max);
+	unsigned consume_src = min(src->bi_iter.bi_size, c->opts.encoded_extent_max);
 	/* Don't generate a bigger output than input: */
-	dst->bi_iter.bi_size = min(dst->bi_iter.bi_size, src->bi_iter.bi_size);
+	unsigned consume_dst = min(dst->bi_iter.bi_size, consume_src);
 
-	compression_type =
-		__bio_compress(c, dst, dst_len, src, src_len,
-			       (union bch_compression_opt){ .value = compression_opt });
+	swap(dst->bi_iter.bi_size, consume_dst);
+	swap(src->bi_iter.bi_size, consume_src);
 
-	dst->bi_iter.bi_size = orig_dst;
-	src->bi_iter.bi_size = orig_src;
+	*src_len = src->bi_iter.bi_size;
+	*dst_len = dst->bi_iter.bi_size;
+
+	struct bbuf dst_data __cleanup(bbuf_exit) = bio_map_or_bounce(c, dst, WRITE);
+	struct bbuf src_data __cleanup(bbuf_exit) = bounce_source
+		? bio_bounce(c, src, src->bi_iter, READ)
+		: bio_map_or_bounce(c, src, READ);
+
+	unsigned compression_type =
+		bch2_compress(c,
+			      dst_data.b, dst_len,
+			      src_data.b, src_len,
+			      compression_opt,
+			      write_pos);
+
+	if (compression_type != BCH_COMPRESSION_TYPE_none &&
+	    compression_type != BCH_COMPRESSION_TYPE_incompressible) {
+		if (dst_data.type != BB_none &&
+		    dst_data.type != BB_vmap)
+			memcpy_to_bio(dst, dst->bi_iter, dst_data.b);
+
+		BUG_ON(!*dst_len || *dst_len > dst->bi_iter.bi_size);
+		BUG_ON(!*src_len || *src_len > src->bi_iter.bi_size);
+	}
+
+	swap(dst->bi_iter.bi_size, consume_dst);
+	swap(src->bi_iter.bi_size, consume_src);
 	return compression_type;
 }
 
@@ -557,6 +615,7 @@ static int __bch2_check_set_has_compressed_data(struct bch_fs *c, u64 f)
 	if ((c->sb.features & f) == f)
 		return 0;
 
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 
 	if ((c->sb.features & f) == f)
