@@ -16,11 +16,10 @@
 #include "btree/write_buffer.h"
 
 #include "data/compress.h"
-#include "data/ec.h"
 #include "data/keylist.h"
 #include "data/move.h"
 #include "data/read.h"
-#include "data/reconcile.h"
+#include "data/reconcile/trigger.h"
 #include "data/reflink.h"
 #include "data/write.h"
 
@@ -65,7 +64,7 @@ static void move_write_done(struct bch_write_op *op)
 	atomic_dec(&ctxt->write_ios);
 
 	bch2_data_update_exit(u, op->error);
-	kfree(u);
+	kfree_rcu(u, rcu);
 	closure_put(&ctxt->cl);
 }
 
@@ -201,12 +200,14 @@ void bch2_move_stats_init(struct bch_move_stats *stats, const char *name)
 	scnprintf(stats->name, sizeof(stats->name), "%s", name);
 }
 
+DEFINE_FREE(data_update_free, struct data_update *, if (_T) kfree_rcu(_T, rcu))
+
 static int __bch2_move_extent(struct moving_context *ctxt,
 		     struct move_bucket *bucket_in_flight,
 		     struct btree_iter *iter,
 		     struct bkey_s_c k,
-		     struct bch_inode_opts io_opts,
-		     struct data_update_opts data_opts)
+		     struct bch_inode_opts *io_opts,
+		     struct data_update_opts *data_opts)
 {
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
@@ -215,7 +216,7 @@ static int __bch2_move_extent(struct moving_context *ctxt,
 	if (ctxt->stats)
 		ctxt->stats->pos = BBPOS(iter->btree_id, iter->pos);
 
-	struct data_update *u __free(kfree) =
+	struct data_update *u __free(data_update_free) =
 		allocate_dropping_locks(trans, ret, kzalloc(sizeof(struct data_update), _gfp));
 	if (!u && !ret)
 		ret = bch_err_throw(c, ENOMEM_move_extent);
@@ -223,7 +224,7 @@ static int __bch2_move_extent(struct moving_context *ctxt,
 		return ret;
 
 	ret = bch2_data_update_init(trans, iter, ctxt, u, ctxt->wp,
-				    &io_opts, data_opts, iter->btree_id, k);
+				    io_opts, *data_opts, iter->btree_id, k);
 	if (ret)
 		return bch2_err_matches(ret, BCH_ERR_data_update_done) ? 0 : ret;
 
@@ -264,17 +265,62 @@ static int __bch2_move_extent(struct moving_context *ctxt,
 			   bkey_start_pos(k.k),
 			   iter->btree_id, k, 0,
 			   NULL,
-			   BCH_READ_last_fragment,
-			   data_opts.read_dev);
+			   data_opts->read_flags|BCH_READ_last_fragment,
+			   data_opts->read_dev);
 	u = NULL;
 	return 0;
 }
 
 int bch2_move_extent(struct moving_context *ctxt,
 		     struct move_bucket *bucket_in_flight,
-		     struct per_snapshot_io_opts *snapshot_io_opts,
-		     move_pred_fn pred, void *arg,
+		     struct bch_inode_opts *opts,
+		     struct data_update_opts *data_opts,
 		     struct btree_iter *iter, unsigned level, struct bkey_s_c k)
+{
+	struct btree_trans *trans = ctxt->trans;
+	struct bch_fs *c = trans->c;
+	int ret = 0;
+
+	if (data_opts->type == BCH_DATA_UPDATE_scrub &&
+	    !bch2_dev_idx_is_online(c, data_opts->read_dev))
+		return bch_err_throw(c, device_offline);
+
+	if (!bkey_is_btree_ptr(k.k))
+		ret = __bch2_move_extent(ctxt, bucket_in_flight, iter, k, opts, data_opts);
+	else if (data_opts->type != BCH_DATA_UPDATE_scrub) {
+		if (data_opts->type != BCH_DATA_UPDATE_copygc)
+			try(bch2_can_do_data_update(trans, opts, data_opts, k, NULL));
+
+		enum bch_trans_commit_flags commit_flags = data_opts->commit_flags;
+		if ((commit_flags & BCH_WATERMARK_MASK) == BCH_WATERMARK_copygc)
+			commit_flags = btree_update_set_watermark_hipri(commit_flags);
+
+		ret = bch2_btree_node_rewrite_pos(trans, iter->btree_id, level, k.k->p,
+						  data_opts->target,
+						  data_opts->commit_flags,
+						  data_opts->write_flags);
+	} else
+		ret = bch2_btree_node_scrub(trans, iter->btree_id, level, k, data_opts->read_dev);
+
+	if (bch2_err_matches(ret, ENOMEM)) {
+		/* memory allocation failure, wait for some IO to finish */
+		bch2_move_ctxt_wait_for_io(ctxt);
+		ret = bch_err_throw(c, transaction_restart_nested);
+	}
+
+	if (!bch2_err_matches(ret, BCH_ERR_transaction_restart) && ctxt->stats)
+		atomic64_add(!bkey_is_btree_ptr(k.k)
+			     ? k.k->size
+			     : c->opts.btree_node_size >> 9, &ctxt->stats->sectors_seen);
+
+	return ret;
+}
+
+static int bch2_move_extent_pred(struct moving_context *ctxt,
+				 struct move_bucket *bucket_in_flight,
+				 struct per_snapshot_io_opts *snapshot_io_opts,
+				 move_pred_fn pred, void *arg,
+				 struct btree_iter *iter, unsigned level, struct bkey_s_c k)
 {
 	if (!bkey_extent_is_direct_data(k.k))
 		return 0;
@@ -310,35 +356,7 @@ int bch2_move_extent(struct moving_context *ctxt,
 	if (ret <= 0)
 		return ret;
 
-	if (data_opts.type == BCH_DATA_UPDATE_scrub &&
-	    !bch2_dev_idx_is_online(c, data_opts.read_dev))
-		return bch_err_throw(c, device_offline);
-
-	if (!bkey_is_btree_ptr(k.k))
-		ret = __bch2_move_extent(ctxt, bucket_in_flight, iter, k, opts, data_opts);
-	else if (data_opts.type != BCH_DATA_UPDATE_scrub) {
-		struct bch_devs_list devs_have = bch2_data_update_devs_keeping(c, &data_opts, k);
-
-		if (data_opts.type != BCH_DATA_UPDATE_copygc)
-			try(bch2_can_do_write(c, &opts, &data_opts, k, &devs_have));
-
-		ret = bch2_btree_node_rewrite_pos(trans, iter->btree_id, level, k.k->p,
-						  data_opts.target, 0, data_opts.write_flags);
-	} else
-		ret = bch2_btree_node_scrub(trans, iter->btree_id, level, k, data_opts.read_dev);
-
-	if (bch2_err_matches(ret, ENOMEM)) {
-		/* memory allocation failure, wait for some IO to finish */
-		bch2_move_ctxt_wait_for_io(ctxt);
-		ret = bch_err_throw(c, transaction_restart_nested);
-	}
-
-	if (!bch2_err_matches(ret, BCH_ERR_transaction_restart) && ctxt->stats)
-		atomic64_add(!bkey_is_btree_ptr(k.k)
-			     ? k.k->size
-			     : c->opts.btree_node_size >> 9, &ctxt->stats->sectors_seen);
-
-	return ret;
+	return bch2_move_extent(ctxt, bucket_in_flight, &opts, &data_opts, iter, level, k);
 }
 
 int bch2_move_ratelimit(struct moving_context *ctxt)
@@ -393,62 +411,41 @@ int bch2_move_data_btree(struct moving_context *ctxt,
 {
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
-	struct bkey_s_c k;
-	int ret = 0;
 
 	CLASS(per_snapshot_io_opts, snapshot_io_opts)(c);
-	CLASS(btree_iter_uninit, iter)(trans);
 
 	if (ctxt->stats) {
 		ctxt->stats->data_type	= BCH_DATA_user;
 		ctxt->stats->pos	= BBPOS(btree_id, start);
 	}
 
-retry_root:
-	bch2_trans_begin(trans);
-
-	if (level == bch2_btree_id_root(c, btree_id)->level + 1) {
-		bch2_trans_node_iter_init(trans, &iter, btree_id, start, 0, level - 1,
-					  BTREE_ITER_prefetch|
-					  BTREE_ITER_not_extents|
-					  BTREE_ITER_all_snapshots);
-		struct btree *b = bch2_btree_iter_peek_node(&iter);
-		ret = PTR_ERR_OR_ZERO(b);
-		if (ret)
-			goto root_err;
-
-		if (b != btree_node_root(c, b))
-			goto retry_root;
-
-		if (btree_node_fake(b))
-			return 0;
-
-		k = bkey_i_to_s_c(&b->key);
-		ret = bch2_move_extent(ctxt, NULL, &snapshot_io_opts, pred, arg, &iter, level, k);
-root_err:
-		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
-			goto retry_root;
+	try(for_btree_root_key_at_level(trans, iter, btree_id, level, k, ({
+		int ret = bch2_move_extent_pred(ctxt, NULL, &snapshot_io_opts, pred, arg, &iter, level, k);
 		if (bch2_err_matches(ret, BCH_ERR_data_update_fail))
 			ret = 0; /* failure for this extent, keep going */
 		WARN_ONCE(ret &&
+			  !bch2_err_matches(ret, BCH_ERR_transaction_restart) &&
+			  !bch2_err_matches(ret, EAGAIN) &&
 			  !bch2_err_matches(ret, EROFS) &&
 			  !bch2_err_matches(ret, EIO),
 			  "unhandled error from move_extent: %s", bch2_err_str(ret));
-		return ret;
-	}
+		ret;
+	})));
 
-	bch2_trans_node_iter_init(trans, &iter, btree_id, start, 0, level,
-				  BTREE_ITER_prefetch|
-				  BTREE_ITER_not_extents|
-				  BTREE_ITER_all_snapshots);
+	bch2_trans_begin(trans);
+	CLASS(btree_node_iter, iter)(trans, btree_id, start, 0, level,
+				     BTREE_ITER_prefetch|
+				     BTREE_ITER_not_extents|
+				     BTREE_ITER_all_snapshots);
 
 	if (ctxt->rate)
 		bch2_ratelimit_reset(ctxt->rate);
 
+	int ret = 0;
 	while (!(ret = bch2_move_ratelimit(ctxt))) {
 		bch2_trans_begin(trans);
 
-		k = bch2_btree_iter_peek(&iter);
+		struct bkey_s_c k = bch2_btree_iter_peek(&iter);
 		if (!k.k)
 			break;
 
@@ -467,7 +464,7 @@ root_err:
 		if (!bkey_extent_is_direct_data(k.k))
 			goto next_nondata;
 
-		ret = bch2_move_extent(ctxt, NULL, &snapshot_io_opts, pred, arg, &iter, level, k);
+		ret = bch2_move_extent_pred(ctxt, NULL, &snapshot_io_opts, pred, arg, &iter, level, k);
 		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
 			continue;
 		if (bch2_err_matches(ret, BCH_ERR_data_update_fail))
@@ -475,7 +472,8 @@ root_err:
 		if (bch2_err_matches(ret, EROFS) ||
 		    bch2_err_matches(ret, EIO)) /* topology error, btree node read error */
 			break;
-		WARN_ONCE(ret, "unhandled error from move_extent: %s", bch2_err_str(ret));
+		WARN_ONCE(ret && !bch2_err_matches(ret, EAGAIN),
+			  "unhandled error from move_extent: %s", bch2_err_str(ret));
 next_nondata:
 		if (!bch2_btree_iter_advance(&iter))
 			break;
@@ -495,7 +493,6 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 {
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
-	bool is_kthread = current->flags & PF_KTHREAD;
 	struct bkey_s_c k;
 	u64 check_mismatch_done = bucket_start;
 	int ret = 0;
@@ -527,9 +524,6 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 		return ret;
 
 	while (!(ret = bch2_move_ratelimit(ctxt))) {
-		if (is_kthread && kthread_should_stop())
-			break;
-
 		bch2_trans_begin(trans);
 
 		k = bch2_btree_iter_peek(&bp_iter);
@@ -537,7 +531,7 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
 			continue;
 		if (ret)
-			return ret;
+			break;
 
 		if (!k.k || bkey_gt(k.k->p, bp_end))
 			break;
@@ -557,7 +551,7 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 		struct bkey_s_c_backpointer bp = bkey_s_c_to_backpointer(k);
 
 		if (ctxt->stats)
-			ctxt->stats->offset = bp.k->p.offset >> MAX_EXTENT_COMPRESS_RATIO_SHIFT;
+			ctxt->stats->offset = bp.k->p.offset >> c->sb.extent_bp_shift;
 
 		if (!(data_types & BIT(bp.v->data_type)) ||
 		    (!bp.v->level && bp.v->btree_id == BTREE_ID_stripes)) {
@@ -571,22 +565,23 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
 			continue;
 		if (ret)
-			return ret;
+			break;
 		if (!k.k) {
 			bch2_btree_iter_advance(&bp_iter);
 			continue;
 		}
 
-		ret = bch2_move_extent(ctxt, bucket_in_flight, NULL, pred, arg, &iter, bp.v->level, k);
+		ret = bch2_move_extent_pred(ctxt, bucket_in_flight, NULL, pred, arg, &iter, bp.v->level, k);
 
 		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
 			continue;
 		if (bch2_err_matches(ret, BCH_ERR_data_update_fail))
 			ret = 0; /* failure for this extent, keep going */
-		if (bch2_err_matches(ret, EROFS) ||
+		if (bch2_err_matches(ret, EAGAIN) ||
+		    bch2_err_matches(ret, EROFS) ||
 		    bch2_err_matches(ret, EIO) ||
 		    bch2_err_matches(ret, BCH_ERR_device_offline))
-			return ret;
+			break;
 		WARN_ONCE(ret, "unhandled error from move_extent: %s", bch2_err_str(ret));
 		bch2_btree_iter_advance(&bp_iter);
 	}
@@ -595,6 +590,7 @@ static int __bch2_move_data_phys(struct moving_context *ctxt,
 		bch2_check_bucket_backpointer_mismatch(trans, ca, check_mismatch_done++,
 						       copygc, &last_flushed);
 
+	bch_err_fn(c, ret);
 	return ret;
 }
 
@@ -638,11 +634,11 @@ static int evacuate_bucket_pred(struct btree_trans *trans, void *_arg,
 		if (ptr->dev == arg->bucket.inode &&
 		    (arg->gen < 0 || arg->gen == ptr->gen) &&
 		    !ptr->cached)
-			data_opts->ptrs_rewrite |= BIT(i);
+			data_opts->ptrs_kill |= BIT(i);
 		i++;
 	}
 
-	return data_opts->ptrs_rewrite != 0;
+	return data_opts->ptrs_kill != 0;
 }
 
 int bch2_evacuate_bucket(struct moving_context *ctxt,
@@ -692,6 +688,7 @@ static int scrub_pred(struct btree_trans *trans, void *_arg,
 
 	data_opts->type		= BCH_DATA_UPDATE_scrub;
 	data_opts->read_dev	= arg->migrate.dev;
+	data_opts->read_flags	= BCH_READ_hard_require_read_device;
 	return true;
 }
 
