@@ -6,13 +6,12 @@
  * Copyright 2012 Google, Inc.
  */
 
-#ifndef NO_BCACHEFS_SYSFS
-
 #include "bcachefs.h"
 
 #include "alloc/accounting.h"
 #include "alloc/background.h"
 #include "alloc/buckets.h"
+#include "alloc/discard.h"
 #include "alloc/disk_groups.h"
 #include "alloc/foreground.h"
 #include "alloc/replicas.h"
@@ -29,10 +28,10 @@
 
 #include "data/compress.h"
 #include "data/copygc.h"
-#include "data/ec.h"
+#include "data/ec/create.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
-#include "data/reconcile.h"
+#include "data/reconcile/work.h"
 
 #include "debug/sysfs.h"
 #include "debug/tests.h"
@@ -57,6 +56,7 @@
 #include <linux/blkdev.h>
 #include <linux/sort.h>
 #include <linux/sched/clock.h>
+#include <linux/sysfs.h>
 
 
 #define SYSFS_OPS(type)							\
@@ -404,8 +404,10 @@ SHOW(bch2_fs)
 	if (attr == &sysfs_disk_groups)
 		bch2_disk_groups_to_text(out, c);
 
-	if (attr == &sysfs_alloc_debug)
+	if (attr == &sysfs_alloc_debug) {
 		bch2_fs_alloc_debug_to_text(out, c);
+		bch2_fs_open_buckets_to_text(out, c);
+	}
 
 	if (attr == &sysfs_usage_base)
 		bch2_fs_usage_base_to_text(out, c);
@@ -714,11 +716,14 @@ static ssize_t sysfs_opt_store(struct bch_fs *c,
 		BUG();
 	}
 
-	if (!ca)
-		bch2_opt_set_by_id(&c->opts, id, v);
+	if (changed) {
+		if (!ca) {
+			bch2_opt_set_by_id(&c->opts, id, v);
+			clear_bit(id, c->mount_opts.d);
+		}
 
-	if (changed)
 		bch2_opt_hook_post_set(c, ca, 0, id, v);
+	}
 
 	ret = size;
 err:
@@ -934,5 +939,3 @@ struct attribute *bch2_dev_files[] = {
 	&sysfs_write_refs,
 	NULL
 };
-
-#endif  /* _BCACHEFS_SYSFS_H_ */
