@@ -61,12 +61,6 @@ static inline bool bch2_dev_idx_is_online(struct bch_fs *c, unsigned dev)
 	return ca && bch2_dev_is_online(ca);
 }
 
-static inline bool bch2_dev_is_healthy(struct bch_dev *ca)
-{
-	return bch2_dev_is_online(ca) &&
-		ca->mi.state != BCH_MEMBER_STATE_evacuating;
-}
-
 static inline unsigned dev_mask_nr(const struct bch_devs_mask *devs)
 {
 	return bitmap_weight(devs->d, BCH_SB_MEMBERS_MAX);
@@ -240,6 +234,18 @@ static inline struct bch_dev *bch2_dev_rcu_noerror(struct bch_fs *c, unsigned de
 		: NULL;
 }
 
+static inline bool bch2_dev_bad_or_evacuating_rcu(struct bch_fs *c, unsigned dev)
+{
+	struct bch_dev *ca = bch2_dev_rcu_noerror(c, dev);
+	return !ca || ca->mi.state == BCH_MEMBER_STATE_evacuating;
+}
+
+static inline bool bch2_dev_bad_or_evacuating(struct bch_fs *c, unsigned dev)
+{
+	guard(rcu)();
+	return bch2_dev_bad_or_evacuating_rcu(c, dev);
+}
+
 int bch2_dev_missing_bkey(struct bch_fs *, struct bkey_s_c, unsigned);
 
 void bch2_dev_missing_atomic(struct bch_fs *, unsigned);
